/*
Copyright 2018 Massachusetts Institute of Technology
Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following
conditions are met:
1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in
the documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

Please reference the article:
"Mihkel Pajusalu, Caue Sciascia Borlina, Sara Seager, Shuhei Ono, Tanja Bosak. Open-source sensor for measuring oxygen partial pressures below 100 microbars. Submitted to PLOS ONE (2018).
for attributing.
*/

//Author: Mihkel Pajusalu, MIT

//This code is intended to be used with FLS Mk8 hardware revision
//The system uses Arduino DUE, which needs to be installed using the Arduino IDE board manager
//The board is intended to be programmed and debugged using the programming USB port (the one closer to the cylindrical connector).

//Before using this code, Adafruit_GFX, Adafruit_ST7735, and SdFat need to be installed
//https://github.com/adafruit/Adafruit-GFX-Library
//https://github.com/adafruit/Adafruit-ST7735-Library
//https://github.com/greiman/SdFat

#include <Adafruit_GFX.h>    // Core graphics library
#include <Adafruit_ST7735.h> // Hardware-specific library
#include <SPI.h>
#include <SdFat.h>

#define MAX_INT 2147483647
#define MAX_FLT 3.4028235E+38
#define MIN_FLT -3.4028235E+38

//User-modifiable constants
#define SENSOR_ID "*FLS_Sensor_1 "//Change this per sensor!!! Identifies each physical sensor uniquely

#define NETWORK_SSID "SSID"
#define NETWORK_PSW  "PASSWORD"
#define SERVER_IP    "127.0.0.1"
#define SERVER_PORT 5000


//Measurement points per cycle
#define cycle_points 2048
#define graphing_downsample 8
#define graphing_points (cycle_points/graphing_downsample)


//Amount of cycles to run (10% will be used for warmup, 90% for actual measurement
#define cycles 400

//Measurement modulation (max intensity of the sine wave), this can be anywhere between 0.0 to 1.0,
//larger values improve signal-to-noise ratio significantly, but also reduce diode and dye operational lifetime
//optimal tradeoff seems to be 0.2, which produces reasonable degradation and SNR
#define MEASUREMENT_MODULATION 0.2

//Reference modulation (max intensity of the reference wave)
#define REFERENCE_MODULATION 1.0

//DAC value corresponding to the minimum of sine wave, needs to be adjusted as low as possible, but very low values deform signal
#define DAC_OFFSET 1200


#define background_count 3


#define AMPLIFIER_POWER_ENABLE 50

//ESP8266 pins
#define ESP_8266_RESET_PIN 51
#define ESP_8266_CH_PD_PIN 53

//Pins used by the display
#define TFT_SCLK 13   
#define TFT_MOSI 12   

#define TFT_CS     9
#define TFT_RST    10  
#define TFT_DC     11
#define TFT_POWER  52

#define BUTTON_CONF 1

Adafruit_ST7735 tft = Adafruit_ST7735(TFT_CS, TFT_DC, TFT_MOSI, TFT_SCLK, TFT_RST);



//Global variables and arrays
int n=0;
float p = 3.1415926;

//Whole lenght of read in samples
#define readbuf (cycles*cycle_points)

//Current position in the read buffer and optput buffer
int rp=-1,wave_pos=-1;

//Arrays to store modulation waveforms
int arr[cycle_points],arrforref[cycle_points];
float arrref[cycle_points];

//Arrays for reading data in
int arr2dummy[cycle_points];
//int arr3dummy[cycle_points];
int arr2cycle[cycle_points],arr2bgcycle[cycle_points];
//int arr3cycle[cycle_points];
int background_accumulator[cycle_points],background_samples,averaging_background=0;


//The gain of the programmable gain amplifier
int dyngain=6;

//Buffer for string processing
char buf[256];

//Mode of the user interface
int user_mode=0;
#define MODE_BG 0
#define MODE_GAIN 1
#define MODE_REF 2

#define USER_MODES 3
bool gain_locked=0;
bool reference_led=0;

//Class for handling basic operations with graphs for display
//The pointers for axes are intended to be reused between individual graphs to save memory
//Therefore the memory has to be allocated separately

class GraphArray{
public:
  float *x,*y;
  int len;

  //Set array pointers to use (this just uses pre-existing arrays, the arrays must be pre-created)
  GraphArray(float *X,float *Y,int count)
  {
    x=X;
    y=Y;
    len=count;
  }
  void SetX(float *X,int downsample)
  {
    for(int n=0;n<len;n++)
      x[n]=X[n*downsample];  
    
  }
  void SetY(float *Y,int downsample)
  {
    for(int n=0;n<len;n++)
      y[n]=Y[n*downsample];  
    
  }

  void SetYDifference(float *A,float *B,int downsample)
  {
    for(int n=0;n<len;n++)
      y[n]=A[n*downsample]-B[n*downsample];  
    
  }

  void SetYDifference(float *A,int B,int downsample)
  {
    for(int n=0;n<len;n++)
      y[n]=A[n*downsample]-B;  
    
  }
  
  void SetYDifference(float *A,float B,int downsample)
  {
    for(int n=0;n<len;n++)
      y[n]=A[n*downsample]-B;  
    
  }

  void SetYDifference(int *A,float B,int downsample)
  {
    for(int n=0;n<len;n++)
      y[n]=A[n*downsample]-B;  
    
  }
  
  
  void SetYMult(float *A,float B,int downsample)
  {
    for(int n=0;n<len;n++)
      y[n]=A[n*downsample]*B;     
    
    
  }

  void SetYMult(int *A,float B,int downsample)
  {
    for(int n=0;n<len;n++)
      y[n]=A[n*downsample]*B;     
    
    
  }

  void SetXMult(float *A,float B,int downsample)
  {
    for(int n=0;n<len;n++)
      x[n]=A[n*downsample]*B;     
    
    
  }

  void NormalizeToRangeY(float dmin,float dmax)
  {
    float t=1./(dmax-dmin);
    for(int n=0;n<len;n++)
    {
      
      y[n]=(y[n]-dmin)*t;
    }
    
  }

  void NormalizeToRangeX(float dmin,float dmax)
  {
    float t=1./(dmax-dmin);
    for(int n=0;n<len;n++)
    {
      
      x[n]=(x[n]-dmin)*t;
    }
    
  }


  void GetMinMaxX( float *dmin_,float *dmax_)
  {
    double dmin,dmax;
    
    dmin=MAX_FLT;
    dmax=MIN_FLT;
    for(int n=0;n<len;n++)
    {
      
      //y[n]=meancyc-arr2cycle[n];
      dmax=((x[n]>dmax)?(x[n]):(dmax));
      dmin=((x[n]<dmin)?(x[n]):(dmin));
    }
    *dmin_=dmin;
    *dmax_=dmax;
  }


  void GetMinMaxY( float *dmin_,float *dmax_)
  {
    double dmin,dmax;
    
    dmin=MAX_FLT;
    dmax=MIN_FLT;
    for(int n=0;n<len;n++)
    {
      dmax=((y[n]>dmax)?(y[n]):(dmax));
      dmin=((y[n]<dmin)?(y[n]):(dmin));
    }
    *dmin_=dmin;
    *dmax_=dmax;
  }


  void NormalizeY()
  {
    float dmin,dmax;
    GetMinMaxY(&dmin,&dmax);
    NormalizeToRangeY(dmin,dmax);
 
  }

  void NormalizeX()
  {
    float dmin,dmax;

    GetMinMaxX(&dmin,&dmax);    
    NormalizeToRangeX(dmin,dmax);
  
  }

};

GraphArray *measurement,*background,*output,*fit,*measurement_ellipse,*fit_ellipse;


//Class for handling display functions, meant to be used with the 1.8 inch SainSmart TFT screen
class TFTDisplay
{
private:
  Adafruit_ST7735 *tft;  
public:
  TFTDisplay(Adafruit_ST7735 *tft)
  {
    this->tft=tft;
  }

  void PowerOff()
  {
    pinMode(TFT_POWER,OUTPUT);
    digitalWrite(TFT_POWER,0);
  }

  void PowerOn()
  {
    pinMode(TFT_POWER,OUTPUT);
    digitalWrite(TFT_POWER,1);
  }


    
  int init()
  {  
    tft->initR(INITR_BLACKTAB);   // initialize a ST7735S chip, black tab
    tft->setRotation(2);
 
    tft->fillScreen(ST7735_BLACK);
  }

  void DrawGraph(GraphArray *graph, int xp,int yp,int xw, int yw, uint16_t color)
  {
    int xb,yb;
    float *x=graph->x;
    float *y=graph->y;

    
    for(int n=0;n<graph->len;n++)
    {
      xb=floor(x[n]*(float)xw)+xp;
      yb=floor(y[n]*(float)yw)+yp;

      tft->drawPixel(xb,yb,color);

    }
    
  }


} *Display;

//Simple circular buffer implementation to hold data
class circbuf
{
  public:
  float *cb;
  int cbpos,cblen,cbfill;
  circbuf(int len)
  {
    cb=(float *)malloc(sizeof(float)*len);
    cbpos=0;
    cblen=len;  
    cbfill=0;
  }  

  void insert(float v)
  {
    cb[cbpos]=v; 
    cbpos=(cbpos+1)%cblen;
    cbfill++;
     
    
  }
  int getlen()
  {
    return min(cbfill,cblen);  
    
    
  }
  float mean()
  {
   float m=0;
   int len=this->getlen();
   for(int n=0;n<len;n++)
   {
    m+=cb[n]; 
   }
   if(len>0) 
    return m/len;
   return 0; 
  }
  float stdev()
  {
    float mx=this->mean(),m=0;  
    int len=this->getlen();
    for(int n=0;n<len;n++)
    {
      m+=pow(cb[n]-mx,2); 
    }
    if(len>0) 
      return sqrt(m/len);
    return 0;  
    
  }  
  static float linreg(circbuf *x,circbuf *y)
  {
    float mx=x->mean();
    float my=y->mean();

    float xmxdot=0.,xmxydot=0.,a,b,rmse;
    int len=x->getlen();
    for(int n=0;n<len;n++)
    {
      xmxdot+=pow(x->cb[n]-mx,2);
      xmxydot+=(x->cb[n]-mx)*y->cb[n];
    }
    a=xmxydot/xmxdot;
    b=my-a*mx;
    for(int n=0;n<len;n++)
    {
      rmse+=pow(y->cb[n]-a*x->cb[n]-b,2);
    }
    rmse=sqrt(rmse/len);
    return rmse;
    
  }
};

circbuf *readings,*readingnums;

int gains[]={0,1,2,5,10,20,50,100};


SdFatSoftSpi<8, 6, 7> sd;
const uint8_t spiSpeed = SPI_FULL_SPEED;
// SD card chip select
#define SD_CS 5
SdFile myFile;

//Tests if SD card is usable
int SDInit()
{
  pinMode(SD_CS,OUTPUT);
  digitalWrite(SD_CS,HIGH);
  delay(100);
  if (!sd.begin(SD_CS, spiSpeed)) {
    return -1;
  }

  Serial.print("\nCard successfully initialized.\n"); 
  return 0; 
}

int SDTest()
{
  SDInit();

  //Test file write and read
  if (!myFile.open("test.txt", O_WRITE | O_CREAT)) {
    sd.errorHalt("opening test.txt for write failed");
  }
  myFile.write("File R/W successful");
  myFile.close();
  if (!myFile.open("test.txt", O_READ)) {
    sd.errorHalt("opening test.txt for read failed");
  }
  
  // read from the file until there's nothing else in it:
  int data;
  while ((data = myFile.read()) >= 0) {
    Serial.write(data);
  }
  // close the file:
  myFile.close();
  return 1;
 }


//Class for storing configuration parameters on the SD card
class NonvolatileConfig
{
  public:
  int gain,active_bg,gain_lock;
  NonvolatileConfig()
  {
    SdFile fin;
    char buf[30];
    int good=0;
    SDInit();
    if(!fin.open("config.txt",O_READ))
      {
        //Serial.println("Read failed, no file");  
        this->gain=1;
        this->active_bg=0;
        this->gain_lock=0;
        this->WriteConfig();
      }
      else
      {
        fin.fgets(buf,29);
        Serial.println(buf);
        good+=sscanf(buf,"%d",&active_bg);
        fin.fgets(buf,29);
        Serial.println(buf);
        good+=sscanf(buf,"%d",&gain);
        fin.fgets(buf,29);
        Serial.println(buf);
        good+=sscanf(buf,"%d",&gain_lock);
        fin.close();

        
      }
    if(good==3)
    {
      Serial.println("Configuration read successful"); 
      Serial.print("Gain ");
      Serial.println(gain);
      Serial.print("Gain lock");
      Serial.println(gain_lock);
      
      Serial.print("BG ");
      Serial.println(active_bg); 
    }
    else
    {
      Serial.println("Configuration read not successful"); 
      Serial.print("Gain ");
      Serial.println(gain);
      Serial.print("Gain lock");
      Serial.println(gain_lock);
      Serial.print("BG ");

      Serial.println(active_bg);    
      gain=3;
      active_bg=0;  
    }
    
  }
  void WriteConfig()
  {
    SdFile fin;
    SDInit();
    sd.remove("config.txt");
    fin.open("config.txt",O_WRITE | O_CREAT );
    Serial.println("Writing config");
    Serial.print("BG ");
    Serial.println(active_bg);
    fin.println(active_bg);
    Serial.print("Gain ");
    Serial.println(gain);
    fin.println(gain);
    Serial.print("Gain lock");
    Serial.println(gain_lock);
    fin.println(gain_lock);
    
    fin.close();
    
  }

  
  
};

NonvolatileConfig *nvconf;

//Class for storing background signals for later subtraction
class BackgroundBuffer
{
  public:
  int bgs[background_count][cycle_points];
  int active_bg;
  BackgroundBuffer()
  {
    char fn[30],buf[256];
    int bufn=0;
    SdFile fin;

    active_bg=0;
    Serial.println("Reading background");
    for(int n=0;n<background_count;n++)
    {
      sprintf(fn,"%d.csv",n);
      Serial.println(fn);
      if(!fin.open(fn,O_READ))
      {
        Serial.println("Read failed, no file");  
        if(!fin.open(fn,O_WRITE | O_CREAT ))
        {
          Serial.println("File creation failed");
          for(int k=0;k<cycle_points;k++)
          {
            bgs[n][k]=0;
            fin.println(0);  
          } 
          continue;
        }
        else
        {
          for(int k=0;k<cycle_points;k++)
          {
            bgs[n][k]=0;
            fin.println(0);  
          }  
          fin.close();
        }
      }
      else
      {
        Serial.println("File exists,attempting read");
        int line=0,tv,data;  
        while ((data = fin.read()) >= 0) {
            //Serial.write(data);
            if(data=='\n')
            {
              bgs[n][line]=atoi(buf);
              line+=1;  
              bufn=0;  
            }
            else
            {    
              buf[bufn]=data;
              bufn++;
            }
            if(line==cycle_points)
              break;
        }
        Serial.print("Lines read: ");
        Serial.println(line);
        fin.close();
      }
    }
    
  }
  void GetBG(int *bg)
  {
    SdFile fin;
    
    for(int n=0;n<cycle_points;n++)
    {
      bg[n]=bgs[active_bg][n];

     
    }
    
  }
  void SetBG(int *bg)
  {
    SdFile fin;
    char fn[30];
    sprintf(fn,"%d.csv",active_bg);
    sd.remove(fn);
    fin.open(fn,O_WRITE | O_CREAT );
    for(int n=0;n<cycle_points;n++)
    {
      bgs[active_bg][n]=bg[n];
      fin.print(bg[n]);
      fin.print(" ");
      fin.println(n);
      fin.flush();
     
    }
    fin.println("0 0");
    fin.close();
    
  }

  void ClearBG()
  {
    SdFile fin;
    char fn[30];
    sprintf(fn,"%d.csv",active_bg);
    sd.remove(fn);
    fin.open(fn,O_WRITE | O_CREAT );
    for(int n=0;n<cycle_points;n++)
    {
      bgs[active_bg][n]=0;
      fin.println(0);
      fin.flush();
    }
    fin.println("0 0");
    fin.close();
    
  }
  
};

//Bacground buffer variable
BackgroundBuffer *bgbuf;

void SetPGAGain(int gain)
{
  pinMode(2,OUTPUT);
  pinMode(3,OUTPUT);
  pinMode(4,OUTPUT);
  digitalWrite(4,gain&1);
  digitalWrite(3,gain&2);
  
  digitalWrite(2,gain&4);
  
  
}

void ESP8266_Off()
{
  pmc_enable_periph_clk(ID_PIOC);
  pinMode(ESP_8266_RESET_PIN,OUTPUT);
  digitalWrite(ESP_8266_RESET_PIN,LOW); 
}


void ESP8266_On()
{
  pmc_enable_periph_clk(ID_PIOC);
  pinMode(ESP_8266_RESET_PIN,OUTPUT);
  digitalWrite(ESP_8266_RESET_PIN,HIGH); 
}

void ESP8266_Setup()
{
//Initialize the serial channels
  Serial.println("Starting communication with the ESP8266 module");

  //Turn on the ESP8266 module
  ESP8266_On();
  delay(1000);

  //For communicating with the optional ESP8266 module (has to be configured properly to work)
  Serial1.begin(115200);

  //Configure the ESP8266 module as a transparent COM port over UDP

  //Enter command mode
  Serial1.print("+++");
  delay(1000);
  serialEvent1();
  Serial1.print("+++");
  delay(200);
  serialEvent1();
  Serial.println();

  //Set working mode
  Serial1.println("AT+CWMODE_DEF=1");
  delay(200);
  serialEvent1();
  Serial.println();

  //Set transparent link parameters
  sprintf(buf,"AT+SAVETRANSLINK=1,\"%s\",%d,\"UDP\",%d",SERVER_IP,SERVER_PORT,SERVER_PORT+1);
  Serial1.println(buf);
  Serial.println(buf);
  
  delay(200);
  serialEvent1();
  Serial.println();

  //Set access point name and password
  sprintf(buf,"AT+CWJAP_DEF=\"%s\",\"%s\"",NETWORK_SSID,NETWORK_PSW);
  Serial1.println(buf);
  Serial.println(buf);
  delay(200);
  serialEvent1();
  Serial.println();

  //Reset ESP8266
  ESP8266_Off();
  delay(200);
  ESP8266_On();  
  
}

void setup() {
  //Powering off the electronics to ensure full reset of amplifiers
  pinMode(AMPLIFIER_POWER_ENABLE,OUTPUT);
  digitalWrite(AMPLIFIER_POWER_ENABLE,LOW);

  
  //Turn off ESP8266 to reset it
  ESP8266_Off();
  
  //For communicating with the computer over the Arduino DUE programming port
  Serial.begin(115200); 
  Serial.println("Starting");



  //Ensure that display is powered off, this is needed to reset display and SD card communication
  Display=new TFTDisplay(&tft);
  Display->PowerOff();  

  //Set analog resolutions
  analogWriteResolution(12);
  analogReadResolution(12);


  //Turn off analog outputs
  analogWrite(DAC0,0);
  analogWrite(DAC1,0);

  //Fill modulation arrays with sine waves
  for(int n=0;n<cycle_points;n++)
  {
    //Main modulation array, contains the waveform emitted through the excitation LED
    arr[n]=floor((4095.-DAC_OFFSET)*(sin(2.*PI*n/cycle_points)+1.)*0.5*MEASUREMENT_MODULATION+DAC_OFFSET);
    //Array to be emitted through the reference LED
    arrforref[n]=floor((4095.-DAC_OFFSET)*(sin(2.*PI*n/cycle_points)+1.)*0.5*REFERENCE_MODULATION+DAC_OFFSET);
    
    //Reference arrays for performing single frequency Fourier transform
    arrref[n]=sin(2.*PI*n/cycle_points);
  }

  //Initialize accumulators for storing data
  for(int n=0;n<cycle_points;n++)
  {
    arr2bgcycle[n]=0;//Background
    arr2cycle[n]=0;
  }

  //Create buffer for storing lifetime readings (default length is 10 readings)
  readings=new circbuf(10);
  readingnums=new circbuf(10);

    //Power on display and SD card  
  Display->PowerOn();
  delay(100);

  //Initialize the display
  Display->init();
  tft.print("Powering on..\n");
  
  //Set up buttons for use
  pinMode(15,INPUT_PULLUP);
  pinMode(16,INPUT_PULLUP);
  pinMode(17,INPUT_PULLUP);
  pinMode(14,INPUT_PULLUP);


  
  //Test SD card status
  //All data is sent to the programming port, so refer to that if errors occur
  int sdstatus=SDTest();
  if(sdstatus!=1)
  {
    tft.setCursor(10,100);
    tft.setTextColor(ST7735_RED);
    tft.print("SDCard init failed");  
    
  }
  else
  {
    tft.print("SD Card detected..\n");  
  }
  
  //Load the configuration from the SD card
  nvconf=new NonvolatileConfig();

  //Set active background
  bgbuf->active_bg=nvconf->active_bg;

  //Set gain
  dyngain=nvconf->gain;
  SetPGAGain(dyngain);
  gain_locked=nvconf->gain_lock;  

  //Create the structure for storing backgrounds
  bgbuf=new BackgroundBuffer();
  bgbuf->active_bg=nvconf->active_bg;
  tft.print("SD card data loaded\n");

  //Create arrays for graphing
  Serial.println("Allocating graphing arrays");
  
  measurement=new GraphArray((float *)malloc(sizeof(float)*graphing_points),(float *)malloc(sizeof(float)*graphing_points),graphing_points);
  background=new GraphArray(measurement->x,(float *)malloc(sizeof(float)*graphing_points),graphing_points);
  output=new GraphArray(measurement->x,(float *)malloc(sizeof(float)*graphing_points),graphing_points);
  fit=new GraphArray(measurement->x,(float *)malloc(sizeof(float)*graphing_points),graphing_points);
  measurement_ellipse=new GraphArray((float *)malloc(sizeof(float)*graphing_points),measurement->y,graphing_points);
  fit_ellipse=new GraphArray(measurement_ellipse->x,fit->y,graphing_points);

  tft.print("Memory allocation completed\n");

   
 
  Serial.println("Filling graphing arrays");
  measurement_ellipse->SetXMult(arrref,-5,graphing_downsample);
  measurement_ellipse->NormalizeX();
 

  for(int n=0;n<graphing_points;n++)
  {
    measurement->x[n]=(float)n/(float)graphing_points;
  }
  tft.print("Starting ESP8266 communication\n");
  
  ESP8266_Setup();
  ESP8266_Off();
   
  Display->init();

  if(sdstatus!=1)
  {
    tft.setCursor(10,100);
    tft.setTextColor(ST7735_RED);
    tft.print("SDCard init failed!");  
    
  }
  digitalWrite(AMPLIFIER_POWER_ENABLE,HIGH);
}

int buttons=0,measuring=1,stop_measuring=0,reading_number=0;


#define GRAPH1X 10
#define GRAPH1Y 40
#define GRAPH1W 100
#define GRAPH1H 65

void loop() {


  //If measuring
  if(measuring)
  {
    
    ESP8266_Off();
    
    //digitalWrite(50,HIGH);
    pmc_disable_all_pck();
    pmc_disable_all_periph_clk(); 
    pmc_enable_periph_clk(ID_ADC);
    adc_init(ADC, SystemCoreClock, ADC_FREQ_MAX*2, 3);
    delay(100);
    // Set up the ADC
    
    adc_channel_num_t ulChannel = (adc_channel_num_t)g_APinDescription[A1].ulADCChannelNumber ;
    adc_enable_channel( ADC, ulChannel );
    ADC->ADC_MR |= 0x80;  //set free running mode on ADC

    //Set up the DAC
    pmc_enable_periph_clk(DACC_INTERFACE_ID);
    dacc_reset(DACC_INTERFACE);
    dacc_set_transfer_mode(DACC_INTERFACE, 0);
    dacc_set_power_save(DACC_INTERFACE, 0, 0);
    dacc_set_timing(DACC_INTERFACE, 0x08, 0, 0x10);  
    dacc_set_analog_control(DACC_INTERFACE, DACC_ACR_IBCTLCH0(0x02) |
                                             DACC_ACR_IBCTLCH1(0x02) |
                                             DACC_ACR_IBCTLDACCORE(0x01));
    
    uint32_t chDACC =((reference_led) ? 1 : 0);
    dacc_set_channel_selection(DACC_INTERFACE, chDACC);
    dacc_enable_channel(DACC_INTERFACE, chDACC);
    uint32_t *dac_output_data=(uint32_t *)((reference_led)?(arrforref):(arr));
    
    uint32_t delayer=10000;
    unsigned long t0=micros();
    for(rp=0;rp<readbuf;rp++)
    {
      
      //Calculate position in the modulation waveform
      wave_pos=rp%cycle_points;
       
      //Modulate the DAC using the waveform in the array 
      DACC_INTERFACE->DACC_CDR=dac_output_data[wave_pos];

      delayer=35;
      while(delayer--)__asm__("nop\n\t");
      
      //For the 10% of the cycles store in a dummy buffer (this dummy buffer is required to keep the modulation speed the same between the first 10% and the last 90% 
      if(rp>=readbuf/10)
      {
        arr2cycle[wave_pos]+=ADC->ADC_CDR[6];
      }
      else
      {
        arr2dummy[wave_pos]+=ADC->ADC_CDR[6];   
      }
    }

    unsigned long t1=micros();
    digitalWrite(AMPLIFIER_POWER_ENABLE,LOW);
    adc_disable_all_channel(ADC);

    //Measure MCU temperature
    delayer=35000;
    while(delayer--)__asm__("nop\n\t");
    adc_enable_channel(ADC, ADC_TEMPERATURE_SENSOR);
    adc_enable_ts(ADC);
    adc_start(ADC);
    delayer=35000;
    while(delayer--)__asm__("nop\n\t");
    int temperature_acc= 0;
    for(int n=0;n<128;n++)
    {
      adc_start(ADC);
      delayer=35000;
      while(delayer--)__asm__("nop\n\t");
      temperature_acc+=adc_get_latest_value(ADC);
      
    
    }
    double temperature=(float)temperature_acc/128.;
    adc_disable_channel(ADC, ADC_TEMPERATURE_SENSOR);
    
    pmc_enable_periph_clk(ID_USART0);
    pmc_enable_periph_clk(ID_USART1);
    pmc_enable_periph_clk(ID_PIOD);
    pmc_enable_periph_clk(ID_PIOA);
    pmc_enable_periph_clk(ID_PIOB);
    pmc_enable_periph_clk(ID_PIOC);
    
    
    //Enable ESP module for network data transfer
    ESP8266_On();
      
    int minv=2147483647 ;
    int maxv=0,reaccint=0,imaccint=0,tvi=0,meancyc=0,meanbgcyc=0,currgain;
    bool change_gain=0;
    double mean=0,b=0;
    double reacc=0,rms=0.;
    double imacc=0,tv,lifetime,signal_strength;
    float dmin2,dmax2,dmin3,dmax3,dmin4,dmax4,dmin5,dmax5;
    analogWrite(DAC0,0);
    analogWrite(DAC1,0);


    currgain=gains[dyngain];

    bgbuf->GetBG(arr2bgcycle);

    //Get signal minimum, maximum, and mean
    for(int n=0;n<cycle_points;n++)
    {
      minv=min(minv,arr2cycle[n]);
      maxv=max(maxv,arr2cycle[n]);
      mean+=arr2cycle[n]/cycle_points;
      arr2cycle[n]*=100./currgain;
      arr2cycle[n]-=arr2bgcycle[n];

      meancyc+=arr2cycle[n]/cycle_points;  
      meanbgcyc+=arr2bgcycle[n]/cycle_points;  
    }


    //Perform the single frequency Fourier transform 
    for(int n=0;n<cycle_points;n++)
    {
      tvi=meancyc-arr2cycle[n];
      reacc+=tvi*arrref[n];
      imacc+=tvi*arrref[(n+cycle_points/4)%cycle_points];
    }

    unsigned long delta_t=t1-t0;
    double modulation_frequency=(double)(cycles*1E6)/(delta_t);
    
    lifetime=-((double)imacc/(double)reacc)*1E3/(2.*PI*(modulation_frequency));
    reacc/=cycle_points;
    imacc/=cycle_points;
    
    readings->insert(lifetime);
    readingnums->insert(reading_number++);
    float rmse=readings->linreg(readingnums,readings);
    
    Serial.println("*");


    //Next block handles dwrawing graphs
    //Draw previous graps as black, i.e. clear previous graphs.
    Display->DrawGraph(measurement,GRAPH1X,GRAPH1Y,GRAPH1W,GRAPH1H,ST7735_BLACK);
    Display->DrawGraph(background,GRAPH1X,GRAPH1Y,GRAPH1W,GRAPH1H,ST7735_BLACK);
    Display->DrawGraph(output,GRAPH1X,GRAPH1Y,GRAPH1W,GRAPH1H,ST7735_BLACK);
    Display->DrawGraph(fit,GRAPH1X,GRAPH1Y,GRAPH1W,GRAPH1H,ST7735_BLACK);
    Display->DrawGraph(measurement_ellipse,10,110,100,40,ST7735_BLACK);
    Display->DrawGraph(fit_ellipse,10,110,100,40,ST7735_BLACK);

    measurement->SetYDifference(arr2cycle,meancyc,graphing_downsample);
    background->SetYDifference(arr2bgcycle,meanbgcyc,graphing_downsample);
    output->SetYMult(arrref,-20000,graphing_downsample);

    rms=0;
    double fitval=0.;
    for(int n=0;n<cycle_points;n++)
    {
      //Calculate the fitted waveform
      fitval=(-arrref[n]*reacc-arrref[(n+cycle_points/4)%cycle_points]*imacc)*2;
      fit->y[n/graphing_downsample]=fitval;
      //Calculate the residual and accumulate for RMS calculation
      b=fitval-arr2cycle[n]+meancyc;
      rms+=b*b;
    }
    
    rms=sqrt(rms/cycle_points);
    signal_strength= sqrt(reacc*reacc+imacc*imacc)*PI/(2.);
    //Normalize the groups of lines to be plotted

    //Normalize the time (phase) vs signal strength graphs 
    measurement->GetMinMaxY(&dmin2,&dmax2);
    output->GetMinMaxY(&dmin3,&dmax3);
    fit->GetMinMaxY(&dmin4,&dmax4);
    background->GetMinMaxY(&dmin5,&dmax5);
    dmin2=min(dmin2,min(dmin3,min(dmin4,dmin5)));
    dmax2=max(dmax2,max(dmax3,max(dmax4,dmax5)));
    dmin2=min(dmin2,-dmax2);
    dmax2=-dmin2;

    //Normalize the ellipse graph
    measurement->NormalizeToRangeY(dmin2,dmax2);
    output->NormalizeToRangeY(dmin2,dmax2);
    fit->NormalizeToRangeY(dmin2,dmax2);
    background->NormalizeToRangeY(dmin2,dmax2);
    
    //Draw new data
    Display->DrawGraph(measurement,GRAPH1X,GRAPH1Y,GRAPH1W,GRAPH1H,ST7735_WHITE);
    Display->DrawGraph(background,GRAPH1X,GRAPH1Y,GRAPH1W,GRAPH1H,ST7735_CYAN);
    Display->DrawGraph(output,GRAPH1X,GRAPH1Y,GRAPH1W,GRAPH1H,ST7735_RED);
    Display->DrawGraph(fit,GRAPH1X,GRAPH1Y,GRAPH1W,GRAPH1H,ST7735_GREEN);
    Display->DrawGraph(measurement_ellipse,10,110,100,40,ST7735_RED);
    Display->DrawGraph(fit_ellipse,10,110,100,40,ST7735_GREEN);


    //Print the waveforms to USB serial port (programming port for Arduino DUE)
    Serial.println("*");
    for(int n=0;n<cycle_points;n++)
    {
      Serial.print(n);
      Serial.print(" ");
      Serial.print(arr2cycle[n]);
      Serial.print(" ");
      Serial.println(arr2bgcycle[n]);       
    }

    //Write statistics
    Serial.print(dyngain);
    Serial.print(" ");

    
    
    Serial.print(signal_strength);
    Serial.print(" ");
    Serial.println(temperature);
    //Send modulation frequency
    Serial.print(modulation_frequency);
    Serial.print(" ");
    sprintf(buf,"%.4f",lifetime*1000.);
    Serial.print(buf);
    Serial.print(" ");
    Serial.println(rms);
    Serial.println("!");



    //Write data to screen
    tft.setCursor(20, 10);
    tft.fillRect(20,10,120,10,ST7735_BLACK);
    tft.setTextColor(ST7735_GREEN);
    tft.setTextWrap(false);
    
    sprintf(buf,"%.2f r%.2e",lifetime*1000,rms);
    tft.print(buf);

    tft.setCursor(0, 20);
    tft.fillRect(0,20,130,10,ST7735_BLACK);
    tft.setCursor(10, 20);
    
    sprintf(buf,"%.3f +- %.3f",readings->mean()*1000,rmse*1000);//readings->stdev()*1000*0.632);
    tft.print(buf);

    
    tft.fillRect(0,30,130,10,ST7735_BLACK);
    
    tft.setCursor(10, 30);

    
    int minvav=minv/cycles;
    

    //Read the button states
    if(BUTTON_CONF)
    buttons=(digitalRead(17)^1)+(digitalRead(16)^1)*2+(digitalRead(15)^1)*4+(digitalRead(14)^1)*8;//TO be used with button PCB
    else
    buttons=(digitalRead(16)^1)+(digitalRead(17)^1)*2+(digitalRead(14)^1)*4+(digitalRead(15)^1)*8;//To be used with membrane buttons
    
    if(buttons==1)
    {
      user_mode=((user_mode<USER_MODES-1)?(user_mode+1):(0));  
    }
    if(user_mode==MODE_BG)
    {
      tft.print("MODE BG ");
      //Background accumulation
      //If the button 2 has been pressed, then clean the accumulation buffer and start accumulating
      if(buttons==2)
      {
        if(!averaging_background)
        {
          background_samples=0;
          for(int n=0;n<cycle_points;n++)
          {
            background_accumulator[n]=0;  
            
          }
          
        }
        for(int n=0;n<cycle_points;n++)
        {
          background_accumulator[n]+=arr2cycle[n];  
          
        }
        background_samples++;
        averaging_background=1;
        
    
      }
      else//If the button has been released, average the background to and store it in the background buffer
      {
    
        if(averaging_background)
        {
          for(int n=0;n<cycle_points;n++)
          {
            background_accumulator[n]/=background_samples;  
            
          }
          bgbuf->SetBG(background_accumulator);
          
        }
        averaging_background=0;
      }
    
      //If button 3 is pressed, change the active background
      if(buttons==4)
      {
        bgbuf->active_bg=((bgbuf->active_bg<(background_count-1))?(bgbuf->active_bg+1):0); 
        nvconf->active_bg=bgbuf->active_bg;
        nvconf->gain=dyngain;
        
        nvconf->WriteConfig();
      }
    
      //Use button 4 is pressed, clear the active background to 0
      if(buttons==8)
      {
        bgbuf->ClearBG(); 
      }
    }
    else if(user_mode==MODE_GAIN)
    {
      tft.print("MODE GAIN ");
      
      if(buttons==2)
      {
        gain_locked=!gain_locked;
        nvconf->gain_lock=gain_locked;
        nvconf->WriteConfig();  
      }
      if(buttons==4)
      {
        dyngain=((dyngain<7)?(dyngain+1):dyngain);
        change_gain=1;
      }
      if(buttons==8)
      {
        dyngain=((dyngain>1)?(dyngain-1):dyngain);  
        change_gain=1;
      }      

      
    }
    else if(user_mode==MODE_REF)
    {
      tft.print("MODE REF");
      
      if(buttons==2)
      {
        reference_led=0;  
      }
      if(buttons==4)
      {
        reference_led=1;  
      }
      tft.print(reference_led);  
    }
    
    if(!gain_locked)
    {
      minvav=minv/cycles;
      //Dynamic gain code, it might be possible to optimize this further, but currently this works
      
      if(minvav<300)
      {
        dyngain=((dyngain>1)?(dyngain-1):dyngain);  
        change_gain=1;
      }
      else if(2048-(2048-minvav)*2.5>400)
      {
        dyngain=((dyngain<7)?(dyngain+1):dyngain);  
        change_gain=1;
        
      }
    }
    if(change_gain)
    {
      SetPGAGain(dyngain);
      nvconf->gain=dyngain;
      nvconf->active_bg=bgbuf->active_bg;
      nvconf->WriteConfig();  
    }

    
    //Reset states of accumulation buffers
    wave_pos=-1;
    for(int n=0;n<cycle_points;n++)
    {
      arr2bgcycle[n]=0;
 
      arr2cycle[n]=0;
    }

    //Stop measuring here (needs to be done here to avoid breaking the state maching)
    if(stop_measuring)
    {
      measuring=0;
      
    }
    sprintf(buf,"G%d%c BG%d ",dyngain,(gain_locked)?('L'):(' '),bgbuf->active_bg);
    tft.print(buf);

    digitalWrite(AMPLIFIER_POWER_ENABLE,HIGH);
    
    delay(500);
    //Write the data to the ESP8266 module
    Serial1.println(SENSOR_ID);
    Serial1.print(dyngain);
    Serial1.print(" ");
    Serial1.print(signal_strength);
    Serial1.print(" ");
    Serial1.println(temperature);
    Serial1.print(modulation_frequency);
   
    Serial1.print(" ");
    sprintf(buf,"%.4f",lifetime*1000.);
    Serial1.print(buf);
    Serial1.print(" ");
    Serial1.println(rms);
    Serial1.println("!");
    delay(500);
    
    
  }
  else
  {
    //If not measuring, write 0 to turn off the DACs
    analogWrite(DAC1,0);
    analogWrite(DAC0,0);
  }


}

//Command parsing, very limited at the moment, since not more is required currently
char inbuf[256],ibp=0;
void parseCommand(char *in,int len)
{
  if(in[0]=='T')
  {
    stop_measuring=1;  
  
  }
  if(in[0]=='B')
  {
    stop_measuring=0;
    measuring=1;
    
  }
 
}

//Accept commands from ESP8266
void serialEvent1()
{
   char a;
   

   while(Serial1.available()>0)
   { 

    
    a=Serial1.read();
    Serial.write(a);
    inbuf[ibp]=a;
    ibp=((ibp<254)?(ibp+1):0);
    if(a=='*')
      ibp=0;
    if(a=='\n')
    {
     parseCommand(inbuf,ibp);
     ibp=0;
    }
   }
}
